# -*- coding: utf-8 -*-
'''
Summary
-------

HITRAN database parser 


Routine Listing
---------------

- :func:`~radis.io.hitran.hit2df`
- :func:`~radis.io.hitran.get_molecule`
- :func:`~radis.io.hitran.get_molecule_identifier`
- :func:`~radis.io.hitran.parse_local_quanta`
- :func:`~radis.io.hitran.parse_global_quanta`


-------------------------------------------------------------------------------


'''

from __future__ import print_function, absolute_import, division, unicode_literals

import sys
import pandas as pd
import numpy as np
from collections import OrderedDict
import os
from os.path import exists, splitext
from six.moves import range
from six.moves import zip
import radis
from radis.misc.cache_files import (check_not_deprecated, save_to_hdf,
                                    DeprecatedFileError, LAST_COMPATIBLE_VERSION)


# %% Hitran groups and classes
# As defined in Rothman et al, "The HITRAN 2004 molecular spectroscopic database"
# Tables 3 and 4

# Groups (define local quanta)

HITRAN_GROUP1 = ['H2O', 'O3', 'SO2', 'NO2', 'HNO3', 'H2CO', 'HOCl', 'H2O2', 'COF2',
                 'H2S', 'HO2', 'HCOOH', 'ClONO2', 'HOBr', 'C2H4']
'''str: asymmetric rotors'''

HITRAN_GROUP2 = ['CO2', 'N2O', 'CO', 'HF', 'HCl', 'HBr', 'HI', 'OCS', 'N2', 'HCN',
                 'C2H2', 'NO+']
'''str: diatomic and linear molecules'''

HITRAN_GROUP3 = ['SF6', 'CH4']
'''str: Spherical rotors'''

HITRAN_GROUP4 = ['CH3D', 'CH3Cl', 'C2H6', 'NH3', 'PH3', 'CH3OH']
'''str: symmetric rotors'''

HITRAN_GROUP5 = ['O2']
'''str: Triplet-Sigma ground electronic states'''

HITRAN_GROUP6 = ['NO', 'OH', 'ClO']
'''str: Doublet-Pi ground electronic states'''

# Classes (define global quanta)

HITRAN_CLASS1 = ['CO', 'HF', 'HCl', 'HBr', 'HI', 'N2', 'NO+']
'''str: Diatomic molecules with ? '''

HITRAN_CLASS2 = ['O2']
'''str: Diatomic molecules with different electronic levels'''

HITRAN_CLASS3 = ['NO', 'OH', 'ClO']
'''str: Diatomic molecules with doublet-Pi electronic state'''

HITRAN_CLASS4 = ['N2O', 'OCS', 'HCN']
'''str: Linear triatomic'''

HITRAN_CLASS5 = ['CO2']
'''str: Linear triatomic with large Fermi resonance'''

HITRAN_CLASS6 = ['H2O', 'O3', 'SO2', 'NO2', 'HOCl', 'H2S', 'HO2', 'HOBr']
'''str: Non-linear triatomic'''

HITRAN_CLASS7 = ['C2H2']
'''str: Linear tetratomic'''

HITRAN_CLASS8 = ['NH3', 'PH3']
'''str: Pyramidal tetratomic'''

HITRAN_CLASS9 = ['H2CO', 'H2O2', 'COF2']
'''str: Non-linear tetratomic'''

HITRAN_CLASS10 = ['CH4', 'CH3D', 'CH3Cl', 'C2H6', 'HNO3', 'SF6', 'HCOOH',
                  'ClONO2', 'C2H4', 'CH3OH']
'''str: Pentatomic or greater polyatomic'''

# %% Parsing functions

# General case
columns_2004 = OrderedDict([(
    # name    # format # type  # description                                 # unit
    'id',     ('a2',   int,   'Molecular number', '')), (
    'iso',    ('a1',   int,   'isotope number', '')), (
    'wav',    ('a12',  float, 'vacuum wavenumber', 'cm-1')), (
    'int',    ('a10',  float, 'intensity at 296K', 'cm-1/(molecule/cm-2)', )), (
    'A',      ('a10',  float, 'Einstein A coefficient', 's-1')), (
    'airbrd', ('a5',   float, 'air-broadened half-width at 296K', 'cm-1.atm-1')), (
    'selbrd', ('a5',   float, 'self-broadened half-width at 296K', 'cm-1.atm-1')), (
    'El',     ('a10',  float, 'lower-state energy', 'cm-1')), (
    'Tdpair', ('a4',   float, 'temperature-dependance exponent for Gamma air', '')), (
    'Pshft',  ('a8',   float, 'air pressure-induced line shift at 296K', 'cm-1.atm-1')), (
    'globu',  ('a15',  str,   'electronic and vibrational global upper quanta', '')), (
    'globl',  ('a15',  str,   'electronic and vibrational global lower quanta', '')), (
    'locu',   ('a15',  str,   'electronic and vibrational local upper quanta', '')), (
    'locl',   ('a15',  str,   'electronic and vibrational local lower quanta', '')), (
    'ierr',   ('a6',   str,   'ordered list of indices corresponding to uncertainty estimates of transition parameters', '')), (
    'iref',   ('a12',  str,   'ordered list of reference identifiers for transition parameters', '')), (
    'lmix',   ('a1',   str,   'flag indicating the presence of additional data and code relating to line-mixing', '')), (
    'gp',     ('a7',   float, 'upper state degeneracy', '')), (
    'gpp',    ('a7',   float, 'lower state degeneracy', ''))
])


def check_cache_file(use_cached, fcache, verbose):
    ''' Check cache file:

    First test its existence: function of the value of ``use_cached``:

    - if True, ok
    - if 'regen', delete cache file to regenerate it later. 
    - if 'force', raise an error if file doesnt exist. 

    Then look if it is deprecated:

    - if deprecated, delete it to regenerate later unless 'force' was used
    '''

    # Test existence of file:
    if not use_cached:
        return                   # we dont want a cache file, no need to test it
    elif use_cached == 'regen':
        if exists(fcache):
            os.remove(fcache)
            if verbose:
                print('Deleted h5 cache file : {0}'.format(fcache))
    elif use_cached == 'force':
        if not exists(fcache):
            raise ValueError('Cache file {0} doesnt exist'.format(fcache))
    else:  # use_cached == True
        pass    # just use the file as is

    # If file is still here, test if it is deprecated:
    if exists(fcache):
        if verbose:
            print('Using cache file: {0}'.format(fcache))
        try:
            check_not_deprecated(fcache, metadata={}, current_version=radis.__version__,
                                 last_compatible_version=LAST_COMPATIBLE_VERSION)
        except DeprecatedFileError as err:
            if use_cached == 'force':
                raise
            else:   # delete file to regenerate it in the end of the script
                if verbose:
                    print('File {0} deprecated:\n{1}\nDeleting it!'.format(
                        fcache, str(err)))
                os.remove(fcache)

    return


def hit2df(fname, count=-1, cache=True, verbose=True):
    ''' Convert a HITRAN/HITEMP [1]_ file to a Pandas dataframe 

    Parameters    
    ----------

    fname: str
        HITRAN-HITEMP file name 

    count: int
        number of items to read (-1 means all file)

    cache: boolean, or ``'regen'`` or ``'force'``
        if ``True``, a pandas-readable HDF5 file is generated on first access, 
        and later used. This saves on the datatype cast and conversion and
        improves performances a lot (but changes in the database are not 
        taken into account). If False, no database is used. If ``'regen'``, temp
        file are reconstructed. Default ``True``. 


    Returns
    -------

    df: pandas Dataframe
        dataframe containing all lines and parameters



    References
    ----------


    .. [1] `HITRAN 1996, Rothman et al., 1998 <https://www.sciencedirect.com/science/article/pii/S0022407398000788>`__



    Notes
    -----

    Performances: see CDSD-HITEMP parser


    See Also
    --------
    
    :func:`~radis.io.cdsd.cdsd2df`

    '''

    columns = columns_2004

    # Use cache file if possible
    fcache = splitext(fname)[0]+'.h5'
    check_cache_file(cache, fcache=fcache, verbose=verbose)
    if cache and exists(fcache):
        return pd.read_hdf(fcache, 'df')

    # Detect the molecule by reading the start of the file
    with open(fname) as f:
        mol = get_molecule(int(f.read(2)))

    # %% Start reading the full file

    df = parse_binary_file(fname, columns, count)

    # %% Post processing

    # assert one molecule per database only. Else the groupbase data reading
    # above doesnt make sense
    nmol = len(set(df['id']))
    if nmol == 0:
        raise ValueError('Databank looks empty')
    elif nmol != 1:
        # Crash, give explicity error messages
        try:
            secondline = df.iloc[1]
        except IndexError:
            secondline = ''
        raise ValueError('Multiple molecules in database ({0}). Current '.format(nmol) +
                         'spectral code only computes 1 species at the time. Use MergeSlabs. ' +
                         'Verify the parsing was correct by looking at the first row below: ' +
                         '\n{0}'.format(df.iloc[0])+'\n----------------\nand the second row ' +
                         'below: \n{0}'.format(secondline))

    # dd local quanta attributes, based on the HITRAN group
    df = parse_local_quanta(df, mol)

    # Add global quanta attributes, based on the HITRAN class
    df = parse_global_quanta(df, mol)

    # cached file mode but cached file doesn't exist yet (else we had returned)
    if cache:
        if verbose:
            print('Generating cached file: {0}'.format(fcache))
        try:
            save_to_hdf(df, fcache, metadata={}, version=radis.__version__,
                        key='df', overwrite=True)
        except:
            if verbose:
                print(sys.exc_info())
                print('An error occured in cache file generation. Lookup access rights')
            pass

    return df


# %% Hitran global quanta classes

def _parse_HITRAN_class1(df):
    ''' Diatomic molecules: CO, HF, HCl, HBr, HI, N2, NO+


    Parameters
    ----------

    df: pandas Dataframe
        lines read from a HITRAN-like database


    Notes
    -----

    HITRAN syntax:

    >>>       v
    >>>  13x I2

    '''
    dgu = df['globu'].str.extract(
        '[ ]{13}(?P<vu>[\d ]{2})',
        expand=True)
    dgl = df['globl'].str.extract(
        '[ ]{13}(?P<vl>[\d ]{2})',
        expand=True)
    dgu = dgu.apply(pd.to_numeric)
    dgl = dgl.apply(pd.to_numeric)
    return pd.concat([df, dgu, dgl], axis=1)


def _parse_HITRAN_class2(df):
    ''' Diatomic molecules with different electronic levels: O2


    Parameters
    ----------

    df: pandas Dataframe
        lines read from a HITRAN-like database


    Notes
    -----

    HITRAN syntax:

    >>>      X   v
    >>>  12x A1 I2

    '''
    print('parse_global_quanta not implemented for molecules of this HITRAN class')
    return df


def _parse_HITRAN_class3(df):
    ''' Diatomic molecules with doublet-Pi electronic state: NO, OH, ClO


    Parameters
    ----------

    df: pandas Dataframe
        lines read from a HITRAN-like database


    Notes
    -------

    HITRAN syntax:

    >>>      X i     v1
    >>>  7x A1 A3 2x I2
    '''
    print('parse_global_quanta not implemented for molecules of this HITRAN class')
    return df


def _parse_HITRAN_class4(df):
    ''' Parse linear triatomic class in HITRAN [1]_: N2O, OCS, HCN

    Parameters
    ----------

    df: pandas Dataframe
        lines read from a HITRAN-like database

    Notes
    -----

    HITRAN syntax:

    >>>     v1 v2 l2 v3
    >>>  7x I2 I2 I2 I2

    Note: I2 in regexp: [\d ]{2} 

    References
    ----------

    .. [1] `HITRAN 1996, Rothman et al., 1998 <https://www.sciencedirect.com/science/article/pii/S0022407398000788>`__

    '''
    dgu = df['globu'].str.extract(
        '[ ]{7}(?P<v1u>[\d ]{2})(?P<v2u>[\d ]{2})(?P<l2u>[\d ]{2})(?P<v3u>[\d ]{2})',
        expand=True)
    dgl = df['globl'].str.extract(
        '[ ]{7}(?P<v1l>[\d ]{2})(?P<v2l>[\d ]{2})(?P<l2l>[\d ]{2})(?P<v3l>[\d ]{2})',
        expand=True)
    dgu = dgu.apply(pd.to_numeric)
    dgl = dgl.apply(pd.to_numeric)
    return pd.concat([df, dgu, dgl], axis=1)


def _parse_HITRAN_class5(df):
    ''' Parse linear triatomic with large Fermi resonance in HITRAN [1]_: CO2

    Parameters
    ----------

    df: pandas Dataframe
        lines read from a HITRAN-like database

    Notes
    -----

    HITRAN syntax:

    >>>     v1 v2 l2 v3 r
    >>>  6x I2 I2 I2 I2 I1

    Note: I2 in regexp: [\d ]{2} 

    References
    ----------

    .. [1] `HITRAN 1996, Rothman et al., 1998 <https://www.sciencedirect.com/science/article/pii/S0022407398000788>`__

    '''
    dgu = df['globu'].str.extract(
        '[ ]{6}(?P<v1u>[\d ]{2})(?P<v2u>[\d ]{2})(?P<l2u>[\d ]{2})(?P<v3u>[\d ]{2})(?P<ru>\d)',
        expand=True)
    dgl = df['globl'].str.extract(
        '[ ]{6}(?P<v1l>[\d ]{2})(?P<v2l>[\d ]{2})(?P<l2l>[\d ]{2})(?P<v3l>[\d ]{2})(?P<rl>\d)',
        expand=True)
    dgu = dgu.apply(pd.to_numeric)
    dgl = dgl.apply(pd.to_numeric)
    return pd.concat([df, dgu, dgl], axis=1)


def _parse_HITRAN_class6(df):
    ''' Parse non-linear triatomic in HITRAN [1]_: H2O, O3, SO2, NO2, HOCl, H2S, HO2, HOBr

    Parameters
    ----------

    df: pandas Dataframe
        lines read from a HITRAN-like database

    Notes
    -----

    HITRAN syntax:

    >>>     v1 v2 v3
    >>>  9x I2 I2 I2

    Note: I2 in regexp: [\d ]{2} 

    References
    ----------

    .. [1] `HITRAN 1996, Rothman et al., 1998 <https://www.sciencedirect.com/science/article/pii/S0022407398000788>`__

    '''
    dgu = df['globu'].str.extract(
        '[ ]{9}(?P<v1u>[\d ]{2})(?P<v2u>[\d ]{2})(?P<v3u>[\d ]{2})',
        expand=True)
    dgl = df['globl'].str.extract(
        '[ ]{9}(?P<v1l>[\d ]{2})(?P<v2l>[\d ]{2})(?P<v3l>[\d ]{2})',
        expand=True)
    dgu = dgu.apply(pd.to_numeric)
    dgl = dgl.apply(pd.to_numeric)
    return pd.concat([df, dgu, dgl], axis=1)


def _parse_HITRAN_class7(df):
    ''' Parse linear tetratomic in HITRAN [1]_: C2H2

    Parameters
    ----------

    df: pandas Dataframe
        lines read from a HITRAN-like database

    Notes
    -----

    HITRAN syntax:

    >>>

    References
    ----------

    .. [1] `HITRAN 1996, Rothman et al., 1998 <https://www.sciencedirect.com/science/article/pii/S0022407398000788>`__

    '''
    print('parse_global_quanta not implemented for molecules of this HITRAN class')
    return df


def _parse_HITRAN_class8(df):
    ''' Pyramidal tetratomic in HITRAN [1]_: NH3, PH3


    Parameters
    ----------

    df: pandas Dataframe
        lines read from a HITRAN-like database


    Notes
    -----

    HITRAN syntax:

    >>>

    References
    ----------

    .. [1] `HITRAN 1996, Rothman et al., 1998 <https://www.sciencedirect.com/science/article/pii/S0022407398000788>`__

    '''
    print('parse_global_quanta not implemented for molecules of this HITRAN class')
    return df


def _parse_HITRAN_class9(df):
    ''' Non-linear tetratomic in HITRAN [1]_: H2CO, H2O2, COF2


    Parameters
    ----------

    df: pandas Dataframe
        lines read from a HITRAN-like database


    Notes
    -----

    HITRAN syntax:

    >>>

    References
    ----------

    .. [1] `HITRAN 1996, Rothman et al., 1998 <https://www.sciencedirect.com/science/article/pii/S0022407398000788>`__

    '''
    print('parse_global_quanta not implemented for molecules of this HITRAN class')
    return df


def _parse_HITRAN_class10(df):
    ''' Pentatomic or greater polyatomic in HITRAN [1]_


    Parameters
    ----------

    df: pandas Dataframe
        lines read from a HITRAN-like database


    Notes
    -----

    HITRAN syntax:

    >>>

    References
    ----------

    .. [1] `HITRAN 1996, Rothman et al., 1998 <https://www.sciencedirect.com/science/article/pii/S0022407398000788>`__

    '''
    print('parse_global_quanta not implemented for molecules of this HITRAN class')
    return df

# %% HITRAN Local quanta


def _parse_HITRAN_group1(df):
    ''' 

    Parameters
    ----------

    df: pandas Dataframe
        lines read from a HITRAN-like database


    Notes
    -----

    HITRAN syntax: [1]_


    References
    ----------

    .. [1] `HITRAN 1996, Rothman et al., 1998 <https://www.sciencedirect.com/science/article/pii/S0022407398000788>`__


    '''
    print('parse_local_quanta not implemented for molecules of this HITRAN group')
    return df


def _parse_HITRAN_group2(df):
    ''' 

    Parameters
    ----------

    df: pandas Dataframe
        lines read from a HITRAN-like database


    Notes
    -----

    HITRAN syntax: [1]


    References
    ----------

    .. [1] `HITRAN 1996, Rothman et al., 1998 <https://www.sciencedirect.com/science/article/pii/S0022407398000788>`__


    '''

    dgu = df['locu'].str.extract(
        '[ ]{10}(?P<Fu>.{5})',
        expand=True)
    dgl = df['locl'].str.extract(
        '[ ]{5}(?P<branch>[\S]{1})(?P<jl>[\d ]{3})(?P<sym>.)(?P<Fl>.{5})',
        expand=True)

    #dgl['jl'] = dgl.jl.apply(pd.to_numeric)
    dgl['jl'] = pd.to_numeric(dgl.jl)

    del df['locu']
    del df['locl']

    return pd.concat([df, dgu, dgl], axis=1)


#                # 10X included in Fu
#               'Fu',     ('a15',  str,  'upper state total angular momentum including nuclear spin'  ,''         )),(
#               #'locl',   ('a15',  str,   'electronic and vibrational local lower quanta'  ,''                      )),(
#               'branch', ('a6',   str,     'O, P, Q, R, S branch symbol'                  ,''                      )),(
#               'jl',     ('a3',   int,    'lower state rotational quantum number'         ,''                      )),(
#               'sym',    ('a1',   str,     'symmetry'                                     ,''                      )),(
#               'Fl',     ('a5',   str,   'lower state total angular momentum including nuclear spin', ''         )),(


def _parse_HITRAN_group3(df):
    ''' 

    Parameters
    ----------

    df: pandas Dataframe
        lines read from a HITRAN-like database


    Notes
    -----

    HITRAN syntax [1]_:


    References
    ----------

    .. [1] `HITRAN 1996, Rothman et al., 1998 <https://www.sciencedirect.com/science/article/pii/S0022407398000788>`__


    '''
    print('parse_local_quanta not implemented for molecules of this HITRAN group')
    return df


def _parse_HITRAN_group4(df):
    ''' 

    Parameters
    ----------

    df: pandas Dataframe
        lines read from a HITRAN-like database


    Notes
    -----

    HITRAN syntax [1]_:


    References
    ----------

    .. [1] `HITRAN 1996, Rothman et al., 1998 <https://www.sciencedirect.com/science/article/pii/S0022407398000788>`__


    '''
    print('parse_local_quanta not implemented for molecules of this HITRAN group')
    return df


def _parse_HITRAN_group5(df):
    ''' 

    Parameters
    ----------

    df: pandas Dataframe
        lines read from a HITRAN-like database


    Notes
    -----

    HITRAN syntax [1]_:


    References
    ----------

    .. [1] `HITRAN 1996, Rothman et al., 1998 <https://www.sciencedirect.com/science/article/pii/S0022407398000788>`__


    '''
    print('parse_local_quanta not implemented for molecules of this HITRAN group')
    return df


def _parse_HITRAN_group6(df):
    ''' 

    Parameters
    ----------

    df: pandas Dataframe
        lines read from a HITRAN-like database


    Notes
    -----

    HITRAN syntax [1]_:


    References
    ----------

    .. [1] `HITRAN 1996, Rothman et al., 1998 <https://www.sciencedirect.com/science/article/pii/S0022407398000788>`__


    '''
    print('parse_local_quanta not implemented for molecules of this HITRAN group')
    return df


# %% Reading function


def _format_dtype(dtype):
    ''' Format dtype from specific columns. Crash with hopefully helping error message '''

    try:
        dt = np.dtype([(str(k), c) for k, c in dtype])
        # Note: dtype names cannot be `unicode` in Python2. Hence the str()
    except TypeError:
        # Cant read database. Try to be more explicit for user
        print('Data type')
        print('-'*30)
        for (k, c) in dtype:
            print(str(k), '\t', c)
        print('-'*30)
        raise
    return dt


def _cast_to_dtype(data, dtype):
    ''' Cast array to certain type, crash with hopefull helping error message.
    Return casted data


    Parameters    
    ----------

    data: array to cast

    dtype: (ordered) list of (param, type)

    '''

    dt = _format_dtype(dtype)

    try:
        data = np.array(data, dtype=dt)
    except ValueError:
        try:
            # Cant read database. Try to be more explicit for user
            print('Cant cast data to specific dtype. Trying column by column:')
            print('-'*30)
            for i in range(len(data[0])):
                print(dtype[i], '\t', np.array(data[0][i], dtype=dt[i]))
            print('-'*30)
        except ValueError:
            print('>>> Next param:', dtype[i], '. Value:', data[0][i], '\n')
            raise ValueError(
                'Cant cast data to specific dtype. Tried column by column. See results above')

    return data


def parse_binary_file(fname, columns, count):
    ''' Parse a file under HITRAN ``par`` format. Parsing is done in binary 
    format so it's as fast as possible.

    Parameters
    ----------

    fname: str
        filename

    columns: dict
        list of columns and their format

    count: int
        number of lines to read

    Returns
    -------

    df: pandas DataFrame
        dataframe with lines

    Notes
    -----

    Part common to hit2df and cdsd2df

    '''

    # To be faster, we read file totally in bytes mode with fromfiles. But that
    # requires to properly decode the line return character:

    # problem arise when file was written in an OS and read in another OS (for instance,
    # line return characters are not converted when read from .egg files). Here
    # we read the first line and infer the line return character for it

    # ... Create a dtype with the binary data format and the desired column names
    dtype = [(k, c[0]) for (k, c) in columns.items()]+[('_linereturn', 'a2')]
    # ... _linereturn is to capture the line return symbol. We delete it afterwards
    dt = _format_dtype(dtype)
    data = np.fromfile(fname, dtype=dt, count=1)   # just read the first line

    # get format of line return
    from radis.misc.basics import to_str
    linereturn = to_str(data[0][-1])
    if to_str('\r\n') in linereturn:
        linereturnformat = 'a2'
    elif to_str('\n') in linereturn or to_str('\r') in linereturn:
        linereturnformat = 'a1'
    else:
        raise ValueError(
            'Line return format unknown: {0}. Please update RADIS'.format(linereturn))

    # Now re-read with correct line return character

    # ... Create a dtype with the binary data format and the desired column names
    dtype = [(k, c[0]) for (k, c) in columns.items()] + \
        [('_linereturn', linereturnformat)]
    # ... _linereturn is to capture the line return symbol. We delete it afterwards
    dt = _format_dtype(dtype)
    data = np.fromfile(fname, dtype=dt, count=count)

    # ... Cast to new type
    # This requires to recast all the data already read, but is still the fastest
    # method I found to read a file directly (for performance benchmark see
    # CDSD-HITEMP parser)
    newtype = [c[0] if (c[1] == str) else c[1] for c in columns.values()]
    dtype = list(zip(list(columns.keys()), newtype)) + \
        [('_linereturn', linereturnformat)]
    data = _cast_to_dtype(data, dtype)

    # %% Create dataframe
    df = pd.DataFrame(data.tolist(), columns=list(
        columns.keys())+['_linereturn'])

    # Delete dummy column than handled the line return character
    del df['_linereturn']

    # Update format
    for k, c in columns.items():
        if c[1] == str:
            df[k] = df[k].str.decode("utf-8")

    # Strip whitespaces around PQR columns (due to 2 columns jumped)
    if 'branch' in df:
        df['branch'] = df.branch.str.strip()

    return df


def parse_local_quanta(df, mol):
    '''
    Parameters
    ----------

    df: pandas Dataframe

    mol: str
        molecule name
    '''

    # had some problems with bytes types
    df['locu'] = df.locu.astype(str)
    df['locl'] = df.locl.astype(str)

    if mol in HITRAN_GROUP1:
        df = _parse_HITRAN_group1(df)
    elif mol in HITRAN_GROUP2:
        df = _parse_HITRAN_group2(df)
    elif mol in HITRAN_GROUP3:
        df = _parse_HITRAN_group3(df)
    elif mol in HITRAN_GROUP4:
        df = _parse_HITRAN_group4(df)
    elif mol in HITRAN_GROUP5:
        df = _parse_HITRAN_group5(df)
    elif mol in HITRAN_GROUP6:
        df = _parse_HITRAN_group6(df)
    else:
        raise ValueError('Unknown group for molecule {0}. Cant parse local quanta'.format(
            mol))

    return df


def parse_global_quanta(df, mol):
    ''' 

    Parameters
    ----------

    df: pandas Dataframe

    mol: str
        molecule name
    '''

    # had some problems with bytes types
    df['globu'] = df.globu.astype(str)
    df['globl'] = df.globl.astype(str)

    if mol in HITRAN_CLASS1:
        df = _parse_HITRAN_class1(df)
    elif mol in HITRAN_CLASS2:
        df = _parse_HITRAN_class2(df)
    elif mol in HITRAN_CLASS3:
        df = _parse_HITRAN_class3(df)
    elif mol in HITRAN_CLASS4:
        df = _parse_HITRAN_class4(df)
    elif mol in HITRAN_CLASS5:
        df = _parse_HITRAN_class5(df)
    elif mol in HITRAN_CLASS6:
        df = _parse_HITRAN_class6(df)
    elif mol in HITRAN_CLASS7:
        df = _parse_HITRAN_class7(df)
    elif mol in HITRAN_CLASS8:
        df = _parse_HITRAN_class8(df)
    elif mol in HITRAN_CLASS9:
        df = _parse_HITRAN_class9(df)
    elif mol in HITRAN_CLASS10:
        df = _parse_HITRAN_class10(df)
    else:
        raise ValueError('Unknown class for molecule {0}. Cant parse global quanta'.format(
            mol))

    return df


def get_molecule_identifier(molecule_name):
    '''
    For a given input molecular formula, return the corresponding HITRAN molecule 
    identifier number [1]_.


    Parameters
    ----------
    molecular_formula : str
        The string describing the molecule.


    Returns
    -------
    M : int
        The HITRAN molecular identified number.


    References
    ----------

    .. [1] `HITRAN 1996, Rothman et al., 1998 <https://www.sciencedirect.com/science/article/pii/S0022407398000788>`__

    Function is from from https://github.com/nzhagen/hitran/blob/master/hitran.py

    '''

    trans = {'1': 'H2O',    '2': 'CO2',   '3': 'O3',      '4': 'N2O',
             '5': 'CO',    '6': 'CH4',   '7': 'O2',     '8': 'NO',
             '9': 'SO2',   '10': 'NO2',  '11': 'NH3',    '12': 'HNO3',
             '13': 'OH',   '14': 'HF',   '15': 'HCl',   '16': 'HBr',
             '17': 'HI',    '18': 'ClO',  '19': 'OCS',    '20': 'H2CO',
             '21': 'HOCl', '22': 'N2',   '23': 'HCN',   '24': 'CH3Cl',
             '25': 'H2O2',  '26': 'C2H2', '27': 'C2H6',   '28': 'PH3',
             '29': 'COF2', '30': 'SF6',  '31': 'H2S',   '32': 'HCOOH',
             '33': 'HO2',   '34': 'O',    '35': 'ClONO2', '36': 'NO+',
             '37': 'HOBr', '38': 'C2H4', '39': 'CH3OH', '40': 'CH3Br',
             '41': 'CH3CN', '42': 'CF4',  '43': 'C4H2',   '44': 'HC3N',
             '45': 'H2',   '46': 'CS',   '47': 'SO3'}
    # Invert the dictionary.
    trans = {v: k for k, v in trans.items()}

    return int(trans[molecule_name])


def get_molecule(molecule_id):
    '''
    For a given input molecular identifier, return the corresponding HITRAN 
    molecule name [1]_.


    Parameters    
    ----------

    molecular_id : str
        Hitran identifier of the molecule.


    References
    ----------

    .. [1] `HITRAN 1996, Rothman et al., 1998 <https://www.sciencedirect.com/science/article/pii/S0022407398000788>`__

    '''

    # assert str
    id = '{0}'.format(molecule_id)

    trans = {'1': 'H2O',    '2': 'CO2',   '3': 'O3',      '4': 'N2O',
             '5': 'CO',    '6': 'CH4',   '7': 'O2',     '8': 'NO',
             '9': 'SO2',   '10': 'NO2',  '11': 'NH3',    '12': 'HNO3',
             '13': 'OH',   '14': 'HF',   '15': 'HCl',   '16': 'HBr',
             '17': 'HI',    '18': 'ClO',  '19': 'OCS',    '20': 'H2CO',
             '21': 'HOCl', '22': 'N2',   '23': 'HCN',   '24': 'CH3Cl',
             '25': 'H2O2',  '26': 'C2H2', '27': 'C2H6',   '28': 'PH3',
             '29': 'COF2', '30': 'SF6',  '31': 'H2S',   '32': 'HCOOH',
             '33': 'HO2',   '34': 'O',    '35': 'ClONO2', '36': 'NO+',
             '37': 'HOBr', '38': 'C2H4', '39': 'CH3OH', '40': 'CH3Br',
             '41': 'CH3CN', '42': 'CF4',  '43': 'C4H2',   '44': 'HC3N',
             '45': 'H2',   '46': 'CS',   '47': 'SO3'}

    return trans[id]

# ======================================================
# %% Test


if __name__ == '__main__':
    from radis.test.test_io import test_hitran_parser
    print('Testing HITRAN parsing: ', test_hitran_parser())



# -*- coding: utf-8 -*-
"""
Summary
-------

Runs tests for neq/misc so that they can be accessed by pytest (and hopefully 
the CI test suite)

Examples
--------

Run all tests::

    pytest      # (in command line, in project folder)

Run only fast tests (i.e: tests that a  'fast' label)::

    pytest -m fast

-------------------------------------------------------------------------------

"""

from __future__ import print_function, absolute_import, division, unicode_literals
import pytest


@pytest.mark.fast
def test_config(*args, **kwargs):
    from radis.misc.config import _test
    return _test(*args, **kwargs)


@pytest.mark.fast
def test_utils(*args, **kwargs):
    from radis.misc.utils import _test
    return _test(*args, **kwargs)


def _run_testcases(verbose=True, *args, **kwargs):

    test_config(verbose=verbose, *args, **kwargs)
    test_utils(verbose=verbose, *args, **kwargs)

    return True


if __name__ == '__main__':
    print(('Testing misc.py: ', _run_testcases(verbose=True)))



# -*- coding: utf-8 -*-
"""
Created on Wed Oct 18 15:38:14 2017

@author: erwan

-------------------------------------------------------------------------------


"""

from __future__ import print_function
from __future__ import absolute_import
import sys
from time import time
from six.moves import range


class ProgressBar():
    ''' A console progress-bar

    Example
    -------

    add a progress bar in a loop::

        pb = ProgressBar(N)
        for i, ... in enumerate(...):
            (...)
            pb.update(i)
        pb.done()

    See test in progress_bar.py

    Notes
    -----

    Todo:

    - One day extend for multiprocss with several progress values? 
      https://stackoverflow.com/questions/7392779/is-it-possible-to-print-a-string-at-a-certain-screen-position-inside-idle

    '''

    def __init__(self, N, active=True):
        ''' 
        write to progress bar completion status i/N. 


        Parameters    
        ----------

        N: int
            total number of iterations

        '''
        self.t0 = time()
        self.N = N
        self.active = active

    def set_active(self, active=True):
        ''' Option to activate/deactivate the ProgressBar. Used not to make it 
        appear on small processes (based on a condition) without changing most
        of the code'''

        self.active = active

    def update(self, i):
        ''' 
        write to progress bar completion status i/N. If t0 is not None, also 
        write the time spent


        Parameters    
        ----------

        i: int
            current iteration
        '''

        if not self.active:
            return

        N = self.N
        t0 = self.t0

        if t0 is None:
            msg = '{0:.1f}%'.format(i/N*100)
        else:
            msg = '({0:.0f}s)\t{1:.1f}%'.format(time()-t0, i/N*100)

        if sys.stdout is not None:
            sys.stdout.write('\r'+msg)
            sys.stdout.flush()

    def done(self):

        if not self.active:
            return

        self.update(self.N)
        # make new line
        if sys.stdout is not None:
            sys.stdout.write('\n')
            sys.stdout.flush()


# %% Test
def test_progress_bar(*args, **kwargs):
    ''' Minimal example of a progress bar '''

    from radis.misc.progress_bar import ProgressBar
    from time import sleep
    from numpy.random import rand

    print('Testing progress bar')

    a = 0
    r = list(range(1000))
    N = len(r)
    pb = ProgressBar(N, t0=time())
    for i in r:
        if i % 10:
            pb.update(i)
        a += i
        sleep(rand()*3e-3)
    pb.done()

    return True  # nothing implemented


if __name__ == '__main__':
    test_progress_bar()



# -*- coding: utf-8 -*-
"""
Test query functions

-------------------------------------------------------------------------------

"""

from __future__ import print_function, absolute_import, division, unicode_literals
import pytest
from radis.io.query import fetch_astroquery


# ignored by pytest with argument -m "not needs_connection"
@pytest.mark.needs_connection
def test_fetch_astroquery(verbose=True, *args, **kwargs):
    ''' Test astroquery '''
    fetch_astroquery('CO2', 1, 2200, 2400, verbose=verbose)

    if verbose:
        print('test_fetch_astroquery: no test defined')

    return


# ignored by pytest with argument -m "not needs_connection"
@pytest.mark.needs_connection
def test_fetch_astroquery_empty(verbose=True, *args, **kwargs):
    ''' Test astroquery: get a spectral range where there are no lines'''
    df = fetch_astroquery(2, 1, 25000, 50000, verbose=verbose)  # 200-400 nm

    assert len(df) == 0

    return


def _run_testcases(verbose=True, *args, **kwargs):

    test_fetch_astroquery(verbose=verbose, *args, **kwargs)
    test_fetch_astroquery_empty(verbose=verbose, *args, **kwargs)

    return True


if __name__ == '__main__':
    print('test_query.py: ', _run_testcases(verbose=True))



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
""" Wrapper to fetch line database from HITRAN with Astroquery [1]_ 

Notes
-----

Astroquery [R1]_ is itself based on HAPI [R2]_ 
    

References
----------

.. [R1] `Astroquery <https://astroquery.readthedocs.io>`_ 

.. [R2] `HAPI: The HITRAN Application Programming Interface <http://hitran.org/hapi>`_


-------------------------------------------------------------------------------

"""

from __future__ import absolute_import
from __future__ import print_function
from radis.io.hitran import get_molecule, get_molecule_identifier
from radis.misc import is_float
from astroquery.hitran import read_hitran_file, cache_location, download_hitran
from os.path import join
import numpy as np
import pandas as pd


def fetch_astroquery(molecule, isotope, wmin, wmax, verbose=True):
    ''' Wrapper to Astroquery [1]_ fetch function to download a line database 

    Notes
    -----

    Astroquery [1]_ is itself based on HAPI [2]_ 

    Parameters
    ----------

    molecule: str, or int
        molecule name or identifier

    isotope: int
        isotope number 

    wmin, wmax: float  (cm-1)
        wavenumber min and max 

    Other Parameters
    ----------------

    verbose: boolean
        Default ``True``

    References
    ----------

    .. [1] `Astroquery <https://astroquery.readthedocs.io>`_ 

    .. [2] `HAPI: The HITRAN Application Programming Interface <http://hitran.org/hapi>`_


    '''

    # Check input
    if not is_float(molecule):
        mol_id = get_molecule_identifier(molecule)
    else:
        mol_id = molecule
        molecule = get_molecule(mol_id)
    assert is_float(isotope)

    empty_range = False

    # Download using the astroquery library
    try:
        download_hitran(mol_id, isotope, wmin, wmax)
    except:
        # Maybe there are just no lines for this species in this range
        # In that case we usually end up with errors like:

            # (<class 'Exception'>, Exception('Query failed: 404 Client Error:
            # Not Found for url: http://hitran.org/lbl/api?numax=25000&numin=19000&iso_ids_list=69\n',),
            # <traceback object at 0x7f0967c91708>)

        import sys
        _err_class, _err_details, _err_obj = sys.exc_info()

        if 'Not Found for url:' in str(_err_details):
            # Let's bet it's just that there are no lines in this range
            empty_range = True
            if verbose:
                print(('Not lines for {0} (id={1}), iso={2} between {3:.2f}-{4:.2f}cm-1. '.format(
                    molecule, mol_id, isotope, wmin, wmax)))
        else:
            raise ValueError('An error occured during the download of HITRAN files ' +
                             'for {0} (id={1}), iso={2} between {3:.2f}-{4:.2f}cm-1. '.format(
                                 molecule, mol_id, isotope, wmin, wmax) +
                             'See details of the error below: {0}'.format(_err_details))

    # Rename columns from Astroquery to RADIS format
    rename_columns = {'molec_id': 'id',
                      'local_iso_id': 'iso',
                      'nu': 'wav',
                      'sw': 'int',
                      'a': 'A',
                      'gamma_air': 'airbrd',
                      'gamma_self': 'selbrd',
                      'elower': 'El',
                      'n_air': 'Tdpair',
                      'delta_air': 'Pshft',
                      'global_upper_quanta': 'globu',
                      'global_lower_quanta': 'globl',
                      'local_upper_quanta': 'locu',
                      'local_lower_quanta': 'locl',
                      'line_mixing_flag': 'lmix',
                      'gp': 'gp',
                      'gpp': 'gpp',
                      }

    if not empty_range:
        tbl = read_hitran_file(join(cache_location, molecule+'.data'))
        df = tbl.to_pandas()
        df = df.rename(columns=rename_columns)
    else:
        df = pd.DataFrame(columns=list(rename_columns.values()))

    # Cast type to float64
    cast_type = {'wav': np.float64,
                 'int': np.float64,
                 'A': np.float64,
                 'airbrd': np.float64,
                 'selbrd': np.float64,
                 'El': np.float64,
                 'Tdpair': np.float64,
                 'Pshft': np.float64,
                 }
    for c, typ in cast_type.items():
        df[c] = df[c].astype(typ)

    return df


if __name__ == '__main__':
    from radis.test.io.test_query import _run_testcases
    _run_testcases(verbose=True)



# -*- coding: utf-8 -*-
"""
Functions and constants used in :class:`~radis.spectrum.spectrum.Spectrum` 
object

-------------------------------------------------------------------------------


"""

from __future__ import print_function, absolute_import, division, unicode_literals
from radis.misc.basics import partition

# %% Definitions

# Waverange, wavespaces
WAVENUM_UNITS = ['cm', 'cm-1', 'cm_1', 'wavenumber']
WAVELEN_UNITS = ['nm', 'wavelength']

# Spectral quantities
CONVOLUTED_QUANTITIES = ['radiance', 'transmittance', 'emissivity']
'''list: name of spectral quantities after convolution with slit function'''
NON_CONVOLUTED_QUANTITIES = ['radiance_noslit', 'transmittance_noslit',
                             'emisscoeff', 'absorbance', 'abscoeff',
                             'abscoeff_continuum', 'emissivity_noslit', ]
'''list: name of spectral quantities not convolved with slit function'''

# note: it is hardcoded (and needed) that quantities that are convoluted are
# generated from a non convoluted quantity with the same name + _noslit
for _q in CONVOLUTED_QUANTITIES:
    assert _q+'_noslit' in NON_CONVOLUTED_QUANTITIES

# %% Special names when printing conditions

# These parameters are shown below "Physical Conditions" rather than "Computation
# Parameters
PHYSICAL_PARAMS = ['molecule', 'wavenum_max', 'wavenum_min', 'mole_fraction',
                   'isotope', 'state', 'path_length', 'medium', 'self_absorption',
                   'slit_function_base', 'pressure_mbar',
                   'wavelength_min', 'wavelength_max',
                   'Telec', 'Tvib', 'Trot', 'Tgas', 'vib_distribution', 'rot_distribution',
                   'overpopulation']

# Informative parameters
# ... Parameters that should be saved in the Spectrum objects, but ignored when
# ... comparing two spectra.
# ... should be written here only these parameters that cannot affect the
# ... physical result. In particular, all parameters relative to performance
# ... should be added here.
INFORMATIVE_PARAMS = [
    'db_use_cached',
    'chunksize',
    'calculation_time',
    'lines_calculated',
    'lines_in_continuum',
    'parallel',
    'Nprocs',
    'Ngroups',
]

# %% Util functions


def cast_waveunit(unit, force_match=True):
    ''' Standardize unit formats '''
    if unit in WAVELEN_UNITS:
        return 'nm'
    elif unit in WAVENUM_UNITS:
        return 'cm-1'
    elif force_match:
        raise ValueError('Unknown wavespace unit: {0}. Should be one of {1}'.format(unit,
                                                                                    WAVELEN_UNITS+WAVENUM_UNITS))
    else:
        return unit  # dont convert


def make_up(label):
    ''' Cosmetic changes on label, before plot 


    Parameters    
    ----------

    label: str
    '''

    # Improve units
    label = label.replace('cm_1', 'cm-1')
    label = label.replace('cm-1', 'cm$^{-1}$')
    label = label.replace('m2', 'm$^2$')
    label = label.replace('m3', 'm$^3$')
    label = label.replace('I/I0', 'I/I$_\mathrm{0}$')    # transmittance unit

    # Improve text
    label = label.replace('_noslit', ' (unconvolved)')
    return label


def print_conditions(conditions, units,
                     phys_param_list=PHYSICAL_PARAMS, info_param_list=INFORMATIVE_PARAMS):
    ''' Print all Spectrum calculation parameters 

    Parameters
    ----------

    phys_param_list: list
        These parameters are shown below "Physical Conditions" rather than "Computation
        Parameters

    info_param_list: list
        These parameters are shown below "Information" rather than "Computation 
        Parameters

    '''

    def align(a, space=20):
        ''' fix alignement '''
        return a + ' ' * max(1, (space - len(str(a))))

    def print_param(k):
        v_k = conditions[k]
        # Add extra arguments based on arbitrary conditions
        args = []
        if k in units:
            args.append(units[k])
        # ... fill here for other args

        # Print
        v_k_str = '{0}'.format(v_k)
        if len(v_k_str) > 102:   # cut if too long
            v_k_str = v_k_str[:100] + '...'
        print(' '*2, align(k), v_k_str, *args)

    phys_param, non_phys_param = partition(lambda x: x in phys_param_list,
                                           conditions)

    info_param, non_phys_param = partition(lambda x: x in info_param_list,
                                           non_phys_param)

    print('Physical Conditions')
    print('-'*40)
    for k in sorted(phys_param):
        print_param(k)

    print('Computation Parameters')
    print('-'*40)
    for k in sorted(non_phys_param):
        print_param(k)

    if len(info_param) > 0:
        print('Information')
        print('-'*40)
        for k in sorted(info_param):
            print_param(k)

    print('-'*40)

    # print gas_inp (information on each gas slab) if exists (specifically
    # for Specair output)
    if 'gas_inp' in conditions:
        try:
            for slab in conditions['gas_inp']:
                print('Slab', slab)
                slab.print_conditions()
        except:
            pass

    return None



# -*- coding: utf-8 -*-
"""

Summary
-------

RADIS 

A code to simulate infrared spectra of molecules::

                *(((((((                                       
                 ((((((((((((              ,(((((              
                 ((((((((((((((((/   *((((((((((*              
                  ((((((((((((((((( ((((((((((((               
                      (((((((( (((((((((((((                   
                         *                                      
                       @@  *@@       ..  /@@                   
                  @@&  @@  *@@       @@  /@@  @@%              
              @@  @@&  @@  *@@  @@&  @@  /@@  @@%  @@          
              @@  @@&  @@  *@@  @@&  @@  /@@  @@%  @@          
              @@  @@&  @@  *@@  @@&  @@  /@@  @@%  @@  (@      
         ,@   @@  @@&  @@  *@@  @@&  @@  /@@  @@%  @@          
         @@   @@  @@&  @@  ,.                                  
                                    ,%&&&&&&&&&&&&&&&&&&&
          &&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&      
           &&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&       
             &&&&&&&&&&&&&&&&@@@@@@&@@@&&&@@@&&&&&&&&          
               &&&&&&&&&&&&&&&@@@@@@&&&&&&&&&&&&&&&            
                 &&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&              
                   &&&&&&&&&&&&&&&&&&&&&&&&&&&.                
                       &&&&&&&&&&&&&&&&&&&                     
                               .**.                            
                                &&&,                         
                                 &&         

See Source code [1]_, Documentation [2]_, Package [3]_

Notes
-----

RADIS is nonequilibrium emission and absorption line-by-line code, for use by 
infrared spectroscopic that want to compare line databases, or experimentalist 
that want to fit their experimental line-of-sight spectra.

Written as a general purpose radiative solver, the code is built around the HITRAN, 
HITEMP and CDSD databases for molecules in their electronic ground state. Energy 
levels are read from tabulated databases or calculated from Dunham developments. 
Boltzmann, Treanor, and state specific vibrational distributions can be 
generated. A modular architecture makes it possible to add new species without 
modifications to the core code. Thus far, CO2, CO are featured for non-equilibrium 
calculations, and all species present in the HITRAN database are featured for 
equilibrium calculations. To fit experimental spectra, RADIS includes a line 
survey tool, an interface with a look-up database to improve fitting convergence 
times, and a multi-slab module with a radiative transfer equation solver to 
reproduce line-of-sight experiments. Validation cases against existing spectral 
codes and experimental results from various plasma sources are presented. 

The code will soon be available under under GNU General Public 
License v3.0 

References
----------

.. [1] Source code: `GitHub repository <https://github.com/radis/radis>`__

.. [2] Online Documentation: `Readthedocs.io <https://radis.readthedocs.io/en/latest/?badge=latest>`__

.. [3] Install as a package: `PyPi project <https://pypi.python.org/pypi/radis>`__

"""

from __future__ import absolute_import, division, print_function, unicode_literals

import os
from .misc.utils import getProjectRoot
from .misc.utils import Chdir as _chdir


# %% Debug mode
DEBUG_MODE = False
# change this at runtime with
# >>> radis.DEBUG_MODE = True
# Use the printdbg() function in radis.misc, typically with:
# >>> if __debug__: printdbg(...)
# so that printdbg are removed by the Python preprocessor when running in
# optimize mode:
# >>> python -O *.py

# %% Version


def get_version(verbose=False, add_git_number=True):
    ''' Reads __version.txt__ and retrieve version number
    If ``add_git``, also appends Git commit number if we're on a gitted session '''

    # First get version
    with open(os.path.join(getProjectRoot(), '__version__.txt')) as version_file:
        version = version_file.read().strip()

    # Now get git info
    if add_git_number:
        import subprocess
        import sys
        cd = _chdir(os.path.dirname(__file__))
        try:
            label = subprocess.check_output('git describe')
            label = label.decode().strip()
            label = '-' + label
        except:
            if verbose:
                print("couldnt get git version: {0}".format(sys.exc_info()[1]))
            # probably not a git session. drop
            label = ''
        finally:
            version = version+label
            cd.__del__()

    return version


__version__ = get_version(add_git_number=False)


from .spectrum import *        # Spectrum object
from .io import *              # input / output
from .lbl import *             # line-by-line module
from .los import *             # line-of-sight module
from .phys import *            # conversion functions, blackbody objects
from .tools import *           # slit, database, line survey, etc.
from .test import *            # test
